# update queries
SET = "set"  # sets an document's field to a new value
UNSET = "unset"  # unset a document's field
PUSH = "push"  # pushes an element into a document's array field
POP = "pop"  # removes either the first or last element in an array
INC = "inc"  # increments a document's numerical field



from .MongoError import *
from .FindQueries import *
from .UpdateQueries import *


class Collection:

    # constructor
    def __init__(self, db, document):
        self._collection = db[document]

    # retrieval functions
    # criteria, all, where, id, object_id
    def find_by_criteria(self, criteria: dict) -> list:
        """
        retrieves every entry in the database based on a criteria dictionary
        :param criteria: dictionary of criteria listing
        :rtype list
        :return every element in the database that satisfies the criteria
        """
        return list(self._collection.find(criteria))

    def find_all(self) -> list:
        """
        retrieves every entry in the database
        :rtype list
        :return every element in the database
        """
        return self.find_by_criteria({})

    def find_where(self, key: str, value: any, aggregate=EQ) -> list:
        """
        find entries based on key-value entry
        :param key: criteria key
        :param value: criteria value
        :param aggregate: default equals
        :rtype list
        :return the entries with the associated criteria
        """
        return self.find_by_criteria({key: {f"${aggregate}": value}})

    def find_by_id(self, _id: any) -> dict:
        """
        find an entry based on an object id
        :param _id: the id to enter
        :rtype dict
        :return the entry w/ associated id
        """
        tmp = list(self._collection.find({"_id": _id}))
        if len(tmp) == 0:
            return {}
        return tmp[0]

    # insertion functions
    # normal, all, id,
    def add(self, entity: dict):
        """
        adds an entry to the database with type ObjectId
        :param entity: the object entity to add
        :raises MongoError: if no id KV pair is specified
        """
        if "_id" not in entity.keys():
            raise MongoError("No id specified")
        self._collection.insert_one(entity)

    def add_all(self, entries: list):
        """
        adds multiple entries to the db
        :param entries: the object entity to add
        """
        for e in entries:
            self.add(e)

    def add_by_id(self, _id: any, entity: dict):
        """
        adds an entry to the database with a user-defined id
        :param _id: the new id to add
        :param entity: the object entity to add
        :except Exception: general exception
        :raises MongoError: error propagated error in try block
        """
        try:
            stub = {'_id': _id}
            stub.update(entity)
            self._collection.insert_one(stub)
        except Exception as e:
            raise MongoError(f"Caused by: {e}")

    # removal functions
    # id, criteria, all
    def remove_by_id(self, _id: any):
        """
        removes an entry based on id of any type
        :param _id: the object associated with id to remove
        """
        self._collection.delete_one({"_id": _id})

    def remove_by_criteria(self, criteria: dict):
        """
        removes multiple entries if they satisfy a criteria
        :param criteria: specific criteria we want to remove by
        """
        self._collection.delete_many(criteria)

    def clear(self):
        """
        clears all documents in the database
        """
        self.remove_by_criteria({})

    # update functions
    # id, criteria, all
    def update_by_criteria(self, criteria: dict, key: str, value: any, aggregate=SET):
        """
        updates the first entry with the matching criteria
        :param criteria: the criteria we want to find the documents by
        :param key: attribute name we want to update
        :param value: value to update to/by
        :param aggregate: default set
        :raises MongoError: raised if user tries to modify the _id field
        """
        if key == "_id":
            raise MongoError("You are not allowed to update the object's id")
        updated = {f"${aggregate}": {key: value}}
        self._collection.update_many(criteria, updated)

    def update_by_id(self, _id: any, key: str, value: any, aggregate=SET):
        """
        updates an entries attributes by finding the entry w/ matching id
        :param _id: the id of the entry we want to update
        :param key: attribute name we want to update
        :param value: value to update to/by
        :param aggregate: default set
        https://docs.mongodb.com/manual/reference/operator/aggregation/set/
        """
        self.update_by_criteria({"_id": _id}, key, value, aggregate)

    def update_all(self, key: str, value: any, aggregate=SET):
        """
        updates the all entries in the collection
        :param key: attribute name we want to update
        :param value: value to update to/by
        :param aggregate: default set
        """
        self.update_by_criteria({}, key, value, aggregate)

    # properties functions
    # size, empty, contains id, contains entry
    def size(self) -> int:
        """
        size of collection
        :rtype int
        :return number of elements in the collection
        """
        return self._collection.count_documents({})

    def empty(self) -> bool:
        """
        sees if collection is empty
        :rtype bool
        :return: true if size is equal to 0
        """
        return self.size() == 0

    def contains_id(self, _id: any) -> bool:
        """
        checks if the collection contains an element based on id
        :param _id: the id to search for
        :rtype bool
        :return true if can find by id
        """
        return len(self.find_by_id(_id)) > 0

    def contains_entry(self, entry: dict) -> bool:
        """
        checks if the collection contains an element
        :param entry: what to search for
        :rtype bool
        :return true if can find by entry that contains criteria
        """
        return len(self.find_by_criteria(entry)) > 0



from .Collection import *


# uses auto-increment to generate the id
class AutoIncrementCollection(Collection):

    # helper functions
    def __probe(self, _id: int) -> int:
        count = 0
        while self.contains_id(_id + count):
            count = count + 1
        return count

    # constructor
    def __init__(self, db, document):
        super().__init__(db, document)

    def find_by_id(self, _id: int) -> dict:
        """
        find an entry based on the integer id
        :param _id: the id to enter
        :rtype dict
        :return the entry w/ associated id
        """
        return super().find_by_id(int(_id))

    # insertion functions

    def add(self, entity: dict):
        """
        adds an entry to the database by auto-incrementing
        :param entity: the object entity to add
        """
        if self.empty():
            self.add_by_id(1, entity)
        else:
            index = self.size()
            offset = self.__probe(index)
            self.add_by_id(index + offset, entity)

    def add_all(self, entries: list):
        """
        adds multiple entries to the db
        :param entries: the object entity to add
        """
        for e in entries:
            self.add(e)

    # removal functions

    def remove_by_id(self, _id: int):
        """
        removes an entry based on an id of int type
        :param _id: the object associated with id to remove
        """
        super().remove_by_id(int(_id))

    def update_by_id(self, _id: int, key: str, value: any, aggregate=SET):
        """
        updates an entries attributes
        :param _id: the id as an int of the entry we want to update
        :param key: attribute name we want to update
        :param value: value to update to/by
        :param aggregate: default set

        https://docs.mongodb.com/manual/reference/operator/aggregation/set/
        """
        super().update_by_id(int(_id), key, value, aggregate)

    def contains_id(self, _id: int) -> bool:
        """
        checks if the collection contains an element based on an int id
        :param _id: the id to search for
        :rtype bool
        :return True if can find by id
        """
        return len(self.find_by_id(_id)) > 0



# find queries
EQ = "eq"  # equals
NE = "ne"  # not equals
GT = "gt"  # greater than
LT = "lt"  # less than
GTE = "gte"  # greater than/equals
LTE = "lte"  # less than/equals
IN = "in"  # array contains
NIN = "nin"  # array not contains



from pymongo import *

from .AutoIncrementCollection import *
from .ObjectIdCollection import *
from .Collection import *
from .MongoError import *

DEFAULT = 0
AUTO_INCREMENT = 1
OBJECT_ID = 2


# MongoDB Wrapper Class
class MongoDB:

    # constructor
    def __init__(
            self,
            database: str,
            collections: dict,
            url: str = None,
            host: str = None,
            port: int = None,

    ):
        """
        :param database: database name to use
        :param collections: collections in the database
        :param url: mongodb url
        :param host: connection host
        :param port: connection port
        :raises MongoError: raised in the following cases
                - host/port specified w/o port/host
                - url specified w/ host and/or port
        """
        if host is None and port is None and url is None:
            connection = MongoClient("localhost", 27017)  # default connection
        elif (host is not None and port is not None) and (url is None):
            connection = MongoClient(host, port)  # user-specified connection
        elif (host is None and port is None) and (url is not None):
            connection = MongoClient(url)  # url-based connection
        else:
            raise MongoError("USAGE:\n"
                             "mongo = MongoDB(database=DATABASE, collections=COLLECTION)\n"
                             "or\n"
                             "mongo = MongoDB(host=HOST, port=PORT, database=DATABASE, collections=COLLECTION)\n"
                             "or\n"
                             "mongo = MongoDB(url=URL, database=DATABASE, collections=COLLECTION)")
        db = connection[database]
        self.collection = {}
        for k in collections:  # k -> collection name
            if collections[k] == DEFAULT:
                self.collection[k] = Collection(db, k)
            if collections[k] == AUTO_INCREMENT:
                self.collection[k] = AutoIncrementCollection(db, k)
            if collections[k] == OBJECT_ID:
                self.collection[k] = ObjectIdCollection(db, k)



from .Collection import *
from bson import ObjectId


class ObjectIdCollection(Collection):
    # helper functions
    @staticmethod
    def __unwrap_all(elements: list) -> list:
        ret = []
        for e in elements:
            e["_id"] = str(e["_id"])
            ret.append(e)
        return ret

    @staticmethod
    def __unwrap(element: dict) -> dict:
        ret = element
        ret["_id"] = str(element["_id"])
        return ret

    # constructor
    def __init__(self, db, document):
        super().__init__(db, document)

    def find_by_criteria(self, criteria: dict) -> list:
        """
        retrieves every entry in the database based on a criteria dictionary
        :param criteria: dictionary of criteria listing
        :rtype list
        :return every element in the database that satisfies the criteria w/ unwrapped id
        """
        return self.__unwrap_all(super().find_by_criteria(criteria))

    def find_where(self, key: str, value: any, aggregate=EQ) -> list:
        """
        find entries based on key-value entry
        :param key: criteria key
        :param value: criteria value
        :param aggregate: default equals
        :rtype list
        :return the entries w/ unwrapped id
        """
        return self.__unwrap_all(super().find_where(key, value, aggregate))

    def find_all(self) -> list:
        """
        retrieves every entry in the database
        :rtype list
        :return every element in the database w/ unwrapped id
        """
        return self.__unwrap_all(super().find_all())

    def find_by_id(self, _id: any) -> dict:
        """
        find entry w/ _id  of type ObjectId
        :param _id: string version of object id
        :rtype dict
        :return the entry with the given id
        """
        return self.__unwrap(super().find_by_id(ObjectId(_id)))

    def add(self, entity: dict):
        """
        adds an entry to the database with type ObjectId
        :param entity: the object entity to add
        """
        self._collection.insert_one(entity)

    def add_all(self, entries: list):
        """
        adds multiple entries to the db
        :param entries: the object entity to add
        """
        self._collection.insert_many(entries)

    def remove_by_id(self, _id: str):
        """
        removes an entry based on id of type ObjectId
        :param _id: ObjectId type _id
        """
        super().remove_by_id(ObjectId(_id))

    def update_by_id(self, _id: str, key: str, value: any, aggregate=SET):
        """
        updates an entries attributes by finding the entry w/ matching id
        :param _id: the id of the entry we want to update
        :param key: attribute name we want to update
        :param value: value to update to/by
        :param aggregate: default set
        https://docs.mongodb.com/manual/reference/operator/aggregation/set/
        """
        super().update_by_id(ObjectId(_id), key, value, aggregate)

    def contains_id(self, _id: str) -> bool:
        """
        checks if the collection contains an element based on id of type ObjectId
        :param _id: the id to search for
        :rtype bool
        :return true if can find by id
        """
        return len(self.find_by_id(_id)) > 0
